import datetime
import json
import logging
import os
import re
import sys
from abc import ABC

from dicom_validator import __version__
from dicom_validator.spec_reader.part3_reader import Part3Reader
from dicom_validator.spec_reader.part4_reader import Part4Reader
from dicom_validator.spec_reader.part6_reader import Part6Reader
from dicom_validator.spec_reader.serializer import DefinitionEncoder

from urllib.request import urlretrieve

import html.parser as html_parser


class EditionParser(html_parser.HTMLParser, ABC):
    edition_re = re.compile(r'\d\d\d\d[a-h]')

    def __init__(self):
        html_parser.HTMLParser.__init__(self)
        self._in_anchor = False
        self.editions = []

    def handle_starttag(self, tag, attrs):
        if tag == 'a':
            self._in_anchor = True

    def handle_endtag(self, tag):
        if tag == 'a':
            self._in_anchor = False

    def handle_data(self, data):
        if self._in_anchor and self.edition_re.match(data):
            self.editions.append(data)


class EditionReader(object):
    base_url = 'http://dicom.nema.org/medical/dicom/'
    html_filename = 'editions.html'
    json_filename = 'editions.json'
    iod_info_json = 'iod_info.json'
    module_info_json = 'module_info.json'
    dict_info_json = 'dict_info.json'
    uid_info_json = 'uid_info.json'

    def __init__(self, path):
        self.path = path
        self.logger = logging.getLogger()
        if not self.logger.handlers:
            self.logger.addHandler(logging.StreamHandler(sys.stdout))

    def update_edition(self):
        try:
            self.logger.info('Getting DICOM editions...')
            self.retrieve(os.path.join(self.path, self.html_filename))
            self.write_to_json()
        except BaseException as exception:
            self.logger.warning(u'Failed to get DICOM read_from_html: %s',
                                str(exception))

    def retrieve(self, html_path):
        if not os.path.exists(os.path.dirname(html_path)):
            os.makedirs(os.path.dirname(html_path))
        urlretrieve(self.base_url, html_path)

    def get_editions(self, update=True):
        editions_path = os.path.join(self.path, self.json_filename)
        if os.path.exists(editions_path):
            if update:
                today = datetime.datetime.today()
                modified_date = datetime.datetime.fromtimestamp(
                    os.path.getmtime(editions_path))
                # no need to update the edition dir more than once a month
                update = (today - modified_date).days > 30
            else:
                with open(editions_path) as f:
                    update = not json.load(f)
        else:
            update = True
        if update:
            self.update_edition()
        if os.path.exists(editions_path):
            with open(editions_path) as json_file:
                return json.load(json_file)

    def read_from_html(self):
        html_path = os.path.join(self.path, self.html_filename)
        with open(html_path) as html_file:
            contents = html_file.read()
        parser = EditionParser()
        parser.feed(contents)
        parser.close()
        return parser.editions

    def write_to_json(self):
        editions = self.read_from_html()
        if editions:
            json_path = os.path.join(self.path, self.json_filename)
            with open(json_path, 'w') as json_file:
                json_file.write(json.dumps(editions))

    def get_edition(self, revision):
        """Get the edition matching the revision or None.
        The revision can be the edition name, the year of the edition,
        'current', or 'local'.
        """
        editions = sorted(self.get_editions(revision != 'local'))
        if revision in editions:
            return revision
        if len(revision) == 4:
            for edition in reversed(editions):
                if edition.startswith(revision):
                    return edition
        if revision == 'current' or revision == 'local':
            return editions[-1]

    def is_current(self, revision):
        """Get the edition matching the revision or None.
        The revision can be the edition name, the year of the edition,
        or 'current'.
        """
        if revision is None:
            return True
        editions = sorted(self.get_editions(revision != 'local'))
        if revision in editions:
            return revision == editions[-1]
        if len(revision) == 4:
            return editions[-1].startswith(revision)
        if revision == 'current':
            return True
        return False

    def check_revision(self, revision):
        # none revision is used if an existing path points to the specs
        if revision != 'none':
            revision = self.get_edition(revision)
            if revision:
                return revision, os.path.join(self.path, revision)
            return None, None
        return None, self.path

    def get_chapter(self, revision, chapter, destination, is_current):
        file_path = os.path.join(destination, 'part{:02}.xml'.format(chapter))
        if os.path.exists(file_path):
            return True
        elif not revision:
            print('Chapter {} not present at {}.'.format(chapter, file_path))
            return False
        revision_dir = 'current' if is_current else revision
        url = '{0}{1}/source/docbook/part{2:02}/part{2:02}.xml'.format(
            self.base_url, revision_dir, chapter)
        try:
            print('Downloading DICOM spec {} PS3.{}...'.format(revision,
                                                               chapter))
            urlretrieve(url, file_path)
            return True
        except BaseException as exception:
            print(u'Failed to download {}: {}'.format(url, str(exception)))
            if os.path.exists(file_path):
                try:
                    os.remove(file_path)
                except OSError:
                    print('Failed to remove incomplete file {}.'.format(
                        file_path))
            return False

    @staticmethod
    def load_info(json_path, info_json):
        with open(os.path.join(json_path,
                               info_json)) as info_file:
            return json.load(info_file)

    @classmethod
    def load_dict_info(cls, json_path):
        return cls.load_info(json_path, cls.dict_info_json)

    @classmethod
    def load_uid_info(cls, json_path):
        return cls.load_info(json_path, cls.uid_info_json)

    @classmethod
    def load_iod_info(cls, json_path):
        return cls.load_info(json_path, cls.iod_info_json)

    @classmethod
    def load_module_info(cls, json_path):
        return cls.load_info(json_path, cls.module_info_json)

    @classmethod
    def json_files_exist(cls, json_path):
        for filename in (cls.dict_info_json, cls.module_info_json,
                         cls.iod_info_json, cls.uid_info_json):
            if not os.path.exists(os.path.join(json_path, filename)):
                return False
        return True

    @classmethod
    def dump_description(cls, description):
        return json.dumps(description, sort_keys=True,
                          indent=2, cls=DefinitionEncoder)

    @classmethod
    def create_json_files(cls, docbook_path, json_path):
        print('Creating JSON excerpts from docbook files...')
        part6reader = Part6Reader(docbook_path)
        dict_info = part6reader.data_elements()
        part3reader = Part3Reader(docbook_path, dict_info)
        part4reader = Part4Reader(docbook_path)
        iod_info = part3reader.iod_descriptions()
        chapter_info = part4reader.iod_chapters()
        definition = {}
        for chapter in iod_info:
            if chapter in chapter_info:
                for uid in chapter_info[chapter]:
                    definition[uid] = iod_info[chapter]
        with open(os.path.join(json_path, cls.iod_info_json),
                  'w') as info_file:
            info_file.write(cls.dump_description(definition))
        with open(os.path.join(json_path, cls.module_info_json),
                  'w') as info_file:
            info_file.write(
                cls.dump_description(part3reader.module_descriptions()))
        with open(os.path.join(json_path, cls.dict_info_json),
                  'w') as info_file:
            info_file.write(cls.dump_description(dict_info))
        with open(os.path.join(json_path, cls.uid_info_json),
                  'w') as info_file:
            info_file.write(cls.dump_description(part6reader.all_uids()))
        cls.write_current_version(json_path)
        print('Done!')

    def get_revision(self, revision):
        revision, destination = self.check_revision(revision)
        if destination is None:
            print('DICOM revision {} not found.'.format(revision))
            return

        docbook_path = os.path.join(destination, 'docbook')
        if not os.path.exists(docbook_path):
            os.makedirs(docbook_path)
        json_path = os.path.join(destination, 'json')
        if not os.path.exists(json_path):
            os.makedirs(json_path)

        # download the docbook files
        for chapter in [3, 4, 6]:
            if not self.get_chapter(revision=revision, chapter=chapter,
                                    destination=docbook_path,
                                    is_current=self.is_current(revision)):
                return

        if (not self.json_files_exist(json_path) or
                not self.is_current_version(json_path)):
            self.create_json_files(docbook_path, json_path)
        print('Using DICOM revision {}'.format(revision))
        return destination

    @staticmethod
    def is_current_version(json_path):
        version_path = os.path.join(json_path, 'version')
        if not os.path.exists(version_path):
            return False
        with open(version_path) as f:
            return f.read() >= __version__

    @staticmethod
    def write_current_version(json_path):
        version_path = os.path.join(json_path, 'version')
        with open(version_path, 'w') as f:
            f.write(__version__)



import os
import unittest

from dicom_validator.tests.test_utils import dicom_fixture_path, fixture_path
from dicom_validator.validate_iods import main


class ValidatorTest(unittest.TestCase):

    @unittest.skip('FIXME: Does not work with TextTestRunner')
    def test_validate_sr(self):
        rtdose_path = os.path.join(dicom_fixture_path(), 'rtdose.dcm')
        cmd_line_args = ['-src', fixture_path(), '-r', 'local', rtdose_path]
        with self.assertLogs('validator', level='INFO') as cm:
            main(cmd_line_args)

        output = ''.join(cm.output)
        # regression test for #9
        self.assertNotIn('Unknown SOPClassUID', output)
        self.assertIn("Tag (0008,1070) (Operators' Name) is missing", output)



import json
import logging
import os
import unittest

from dicom_validator.spec_reader.edition_reader import EditionReader

from pydicom.dataset import Dataset

from dicom_validator.tests.test_utils import json_fixture_path
from dicom_validator.validator.iod_validator import IODValidator


class IODValidatorTest(unittest.TestCase):
    """Tests IODValidator.
    Note: some of the fixture data are not consistent with the DICOM Standard.
    """
    iod_specs = None
    module_specs = None

    @classmethod
    def setUpClass(cls):
        with open(os.path.join(json_fixture_path(),
                               EditionReader.iod_info_json)) as info_file:
            cls.iod_specs = json.load(info_file)
        with open(os.path.join(json_fixture_path(),
                               EditionReader.module_info_json)) as info_file:
            cls.module_specs = json.load(info_file)

    def setUp(self):
        super(IODValidatorTest, self).setUp()
        logging.disable(logging.CRITICAL)

    def validator(self, data_set):
        return IODValidator(data_set, self.iod_specs, self.module_specs, None,
                            logging.ERROR)

    @staticmethod
    def new_data_set(tags):
        """ Create a DICOM data set with the given attributes """
        tags = tags or {}
        data_set = Dataset()
        for tag_name, value in tags.items():
            setattr(data_set, tag_name, value)
        data_set.file_meta = Dataset()
        data_set.is_implicit_VR = False
        data_set.is_little_endian = True
        return data_set

    def test_empty_dataset(self):
        data_set = self.new_data_set(tags={})
        validator = self.validator(data_set)
        result = validator.validate()
        self.assertIn('fatal', result)

    def test_invalid_sop_class_id(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.3'
        })
        validator = self.validator(data_set)
        result = validator.validate()
        self.assertIn('fatal', result)

    @staticmethod
    def has_tag_error(messages, module_name, tag_id_string, error_kind):
        if module_name not in messages:
            return False
        for message in messages[module_name]:
            if message.startswith(
                    'Tag {} is {}'.format(tag_id_string, error_kind)):
                return True
        return False

    def test_missing_tags(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.2',  # CT
            'PatientName': 'XXX',
            'PatientID': 'ZZZ',
        })
        validator = self.validator(data_set)
        result = validator.validate()

        self.assertNotIn('fatal', result)
        self.assertIn('CT Image', result)

        # PatientName is set
        self.assertFalse(
            self.has_tag_error(result, 'Patient', '(0010,0010)', 'missing'))
        # PatientSex - type 2, missing
        self.assertTrue(
            self.has_tag_error(result, 'Patient', '(0010,0040)', 'missing'))
        # Clinical Trial Sponsor Name -> type 1, but module usage U
        self.assertFalse(
            self.has_tag_error(result, 'Patient', '(0012,0010)', 'missing'))
        # Patient Breed Description -> type 2C, but no parsable condition
        self.assertFalse(
            self.has_tag_error(result, 'Patient', '(0010,2292)', 'missing'))

    def test_empty_tags(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.2',  # CT
            'PatientName': '',
            'Modality': None
        })
        validator = self.validator(data_set)
        result = validator.validate()

        self.assertNotIn('fatal', result)
        self.assertIn('CT Image', result)
        # Modality - type 1, present but empty
        self.assertTrue(
            self.has_tag_error(result, 'Patient', '(0010,0040)', 'missing'))
        # PatientName - type 2, empty tag is allowed
        self.assertFalse(
            self.has_tag_error(result, 'Patient', '(0010,0010)', 'missing'))

    def test_fulfilled_condition_existing_tag(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.12.1.1',
            # Enhanced X-Ray Angiographic Image
            'CArmPositionerTabletopRelationship': 'YES',
            'SynchronizationTrigger': 'SET',
            'FrameOfReferenceUID': '1.2.3.4.5.6.7.8',
            'PatientName': 'XXX',
            'PatientID': 'ZZZ'
        })
        validator = self.validator(data_set)
        result = validator.validate()

        # Frame Of Reference UID Is and Synchronization Trigger set
        self.assertFalse(
            self.has_tag_error(result, 'Enhanced X-Ray Angiographic Image',
                               '(0020,0052)', 'missing'))
        self.assertFalse(self.has_tag_error(result, 'Synchronization',
                                            '(0018,106A)', 'missing'))

    def test_fulfilled_condition_missing_tag(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.12.1.1',
            # Enhanced X-Ray Angiographic Image
            'CArmPositionerTabletopRelationship': 'YES',
            'PatientName': 'XXX',
            'PatientID': 'ZZZ'
        })
        validator = self.validator(data_set)
        result = validator.validate()

        self.assertTrue(self.has_tag_error(result, 'Frame of Reference',
                                           '(0020,0052)', 'missing'))
        self.assertTrue(self.has_tag_error(result, 'Synchronization',
                                           '(0018,106A)', 'missing'))

    def test_condition_not_met_no_tag(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.12.1.1',
            # Enhanced X-Ray Angiographic Image
            'PatientName': 'XXX',
            'PatientID': 'ZZZ'
        })
        validator = self.validator(data_set)
        result = validator.validate()

        self.assertFalse(self.has_tag_error(result, 'Frame of Reference',
                                            '(0020,0052)', 'missing'))
        self.assertFalse(self.has_tag_error(result, 'Frame of Reference',
                                            '(0020,0052)', 'not allowed'))

    def test_condition_not_met_existing_tag(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.12.1.1',
            # Enhanced X-Ray Angiographic Image
            'FrameOfReferenceUID': '1.2.3.4.5.6.7.8',
            'SynchronizationTrigger': 'SET',
            'PatientName': 'XXX',
            'PatientID': 'ZZZ'
        })
        validator = self.validator(data_set)
        result = validator.validate()

        # Frame Of Reference is allowed, Synchronization Trigger not
        self.assertFalse(self.has_tag_error(result, 'Frame of Reference',
                                            '(0020,0052)', 'missing'))
        self.assertFalse(self.has_tag_error(result, 'Frame of Reference',
                                            '(0020,0052)', 'not allowed'))
        self.assertTrue(self.has_tag_error(result, 'Synchronization',
                                           '(0018,106A)', 'not allowed'))

    def test_and_condition_not_met(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.12.1.1',
            # Enhanced X-Ray Angiographic Image
            'PatientName': 'XXX',
            'PatientID': 'ZZZ',
            'ImageType': 'SECONDARY',
            'CardiacSynchronizationTechnique': 'OTHER',
            'HighRRValue': '123'  # 0018,1082
        })
        validator = self.validator(data_set)
        result = validator.validate()

        # Both Low R-R Value and High R-R Value are not needed but allowed
        self.assertFalse(self.has_tag_error(result, 'Cardiac Synchronization',
                                            '(0018,1081)', 'missing'))
        self.assertFalse(self.has_tag_error(result, 'Cardiac Synchronization',
                                            '(0018,1082)', 'missing'))

    def test_only_one_and_condition_met(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.12.1.1',
            # Enhanced X-Ray Angiographic Image
            'PatientName': 'XXX',
            'PatientID': 'ZZZ',
            'ImageType': 'PRIMARY',
            'CardiacSynchronizationTechnique': 'OTHER',
            'HighRRValue': '123'  # 0018,1082
        })
        validator = self.validator(data_set)
        result = validator.validate()

        # Both Low R-R Value and High R-R Value are not needed but allowed
        self.assertFalse(self.has_tag_error(result, 'Cardiac Synchronization',
                                            '(0018,1081)', 'missing'))
        self.assertFalse(self.has_tag_error(result, 'Cardiac Synchronization',
                                            '(0018,1082)', 'missing'))

    def test_and_condition_met(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.12.1.1',
            # Enhanced X-Ray Angiographic Image
            'PatientName': 'XXX',
            'PatientID': 'ZZZ',
            'ImageType': 'MIXED',
            'CardiacSynchronizationTechnique': 'PROSPECTIVE',
            'HighRRValue': '123'  # 0018,1082
        })
        validator = self.validator(data_set)
        result = validator.validate()

        # Both Low R-R Value and High R-R Value are needed
        self.assertTrue(self.has_tag_error(result, 'Cardiac Synchronization',
                                           '(0018,1081)', 'missing'))
        self.assertFalse(self.has_tag_error(result, 'Cardiac Synchronization',
                                            '(0018,1082)', 'missing'))

    def test_presence_condition_met(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.12.1.1',
            # Enhanced X-Ray Angiographic Image
            'PatientName': 'XXX',
            'PatientID': 'ZZZ',
            'PixelPaddingRangeLimit': '10',
            'PixelDataProviderURL': 'https://dataprovider'
        })
        validator = self.validator(data_set)
        result = validator.validate()

        self.assertTrue(self.has_tag_error(result, 'General Equipment',
                                           '(0028,0120)',
                                           'missing'))  # Pixel Padding Value

    def test_presence_condition_not_met(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.12.1.1',
            # Enhanced X-Ray Angiographic Image
            'PatientName': 'XXX',
            'PatientID': 'ZZZ',
            'PixelPaddingRangeLimit': '10',
        })
        validator = self.validator(data_set)
        result = validator.validate()

        self.assertFalse(self.has_tag_error(result, 'General Equipment',
                                            '(0028,0120)',
                                            'missing'))  # Pixel Padding Value

    def test_greater_condition_met(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.12.1.1',
            # Enhanced X-Ray Angiographic Image
            'PatientName': 'XXX',
            'PatientID': 'ZZZ',
            'SamplesPerPixel': 3
        })
        validator = self.validator(data_set)
        result = validator.validate()

        self.assertTrue(self.has_tag_error(result, 'Image Pixel',
                                           '(0028,0006)',
                                           'missing'))  # Planar configuration

    def test_greater_condition_not_met(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.12.1.1',
            # Enhanced X-Ray Angiographic Image
            'PatientName': 'XXX',
            'PatientID': 'ZZZ',
            'SamplesPerPixel': 1
        })
        validator = self.validator(data_set)
        result = validator.validate()

        self.assertFalse(self.has_tag_error(result, 'Image Pixel',
                                            '(0028,0006)',
                                            'missing'))  # Planar configuration

    def test_points_to_condition_met(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.12.1.1',
            # Enhanced X-Ray Angiographic Image
            'PatientName': 'XXX',
            'PatientID': 'ZZZ',
            'FrameIncrementPointer': 0x00181065
        })
        validator = self.validator(data_set)
        result = validator.validate()

        self.assertTrue(self.has_tag_error(result, 'Cardiac Synchronization',
                                           '(0018,1086)',
                                           'missing'))  # Skip beats

    def test_points_to_condition_not_met(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.12.1.1',
            # Enhanced X-Ray Angiographic Image
            'PatientName': 'XXX',
            'PatientID': 'ZZZ',
            'FrameIncrementPointer': 0x00181055
        })
        validator = self.validator(data_set)
        result = validator.validate()

        self.assertFalse(self.has_tag_error(result, 'Cardiac Synchronization',
                                            '(0018,1086)',
                                            'missing'))  # Skip beats

    def test_condition_for_not_required_tag_cond1_fulfilled(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.12.1.1',
            # Enhanced X-Ray Angiographic Image
            'PatientName': 'XXX',
            'PatientID': 'ZZZ',
            'ImageType': 'ORIGINAL',
            'CardiacSynchronizationTechnique': 'ANY'
        })
        validator = self.validator(data_set)
        result = validator.validate()

        self.assertTrue(self.has_tag_error(result, 'Cardiac Synchronization',
                                           '(0018,9085)',
                                           'missing'))  # Cardiac signal source

    def test_condition_for_not_required_tag_no_cond_fulfilled(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.12.1.1',
            # Enhanced X-Ray Angiographic Image
            'PatientName': 'XXX',
            'PatientID': 'ZZZ',
            'ImageType': 'ORIGINAL',
            'CardiacSynchronizationTechnique': 'NONE',
            'CardiacSignalSource': 'ECG'
        })
        validator = self.validator(data_set)
        result = validator.validate()

        self.assertTrue(self.has_tag_error(
            result, 'Cardiac Synchronization',
            '(0018,9085)',
            'not allowed'))  # Cardiac signal source

    def test_condition_for_not_required_tag_cond2_fulfilled_present(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.12.1.1',
            # Enhanced X-Ray Angiographic Image
            'PatientName': 'XXX',
            'PatientID': 'ZZZ',
            'ImageType': 'DERIVED',
            'CardiacSynchronizationTechnique': 'ANY',
            'CardiacSignalSource': 'ECG'
        })
        validator = self.validator(data_set)
        result = validator.validate()

        self.assertFalse(self.has_tag_error(
            result, 'Cardiac Synchronization',
            '(0018,9085)',
            'not allowed'))  # Cardiac signal source

    def test_condition_for_not_required_tag_cond2_fulfilled_not_present(self):
        data_set = self.new_data_set({
            'SOPClassUID': '1.2.840.10008.5.1.4.1.1.12.1.1',
            # Enhanced X-Ray Angiographic Image
            'PatientName': 'XXX',
            'PatientID': 'ZZZ',
            'ImageType': 'DERIVED',
            'CardiacSynchronizationTechnique': 'ANY'
        })
        validator = self.validator(data_set)
        result = validator.validate()

        self.assertFalse(self.has_tag_error(
            result, 'Cardiac Synchronization',
            '(0018,9085)',
            'missing'))  # Cardiac signal source


if __name__ == '__main__':
    unittest.main()



def tag_name_from_id_string(tag_id_string, dict_info):
    if dict_info and tag_id_string in dict_info:
        return '{} ({})'.format(
            tag_id_string, dict_info[tag_id_string]['name'])
    return tag_id_string


def tag_name_from_id(tag_id, dict_info):
    tag_id_string = '({:04X},{:04X})'.format(
        tag_id // 0x10000, tag_id % 0x10000)
    return tag_name_from_id_string(tag_id_string, dict_info)



import json

from dicom_validator.tag_tools import tag_name_from_id


class Condition(object):
    """ Represents a condition for the presence of a specific tag.

    Attributes:
        type: the type of the related object (tag or module) regarding its
            existence; possible values:
            'U': user defined, e.g. both existence or non-existence
                of the related object is considered legal
            'MN': the object is mandatory if the condition is fulfilled,
                otherwise not
            'MU': the object is mandatory if the condition is fulfilled,
                otherwise is user defined
        tag: the ID of the required tag in the form '(####,####)' or None
        index: the index of the tag for multi-valued tags or 0
        values: a list of values the tag shall have if the condition
            is fulfilled, or None
        operator: the comparison operation used ('=', '<', '>') for the
            value(s), or None

    """

    def __init__(self, ctype=None, operator=None, tag=None, index=0,
                 values=None):

        self.type = ctype
        self.operator = operator
        self.tag = tag
        self.index = index
        self.values = values
        self.and_conditions = []
        self.or_conditions = []
        self.other_condition = None

    def read(self, json_string):
        condition_dict = json.loads(json_string)
        self.read_condition(condition_dict, self)

    @classmethod
    def read_condition(cls, condition_dict, condition=None):
        condition = condition or Condition()
        condition.type = condition_dict.get('type')
        condition.operator = condition_dict.get('op')
        condition.tag = condition_dict.get('tag')
        condition.index = int(condition_dict.get('index', '0'))
        condition.values = condition_dict.get('values')
        and_list = condition_dict.get('and', [])
        condition.and_conditions = [
            cls.read_condition(cond) for cond in and_list]
        or_list = condition_dict.get('or', [])
        condition.or_conditions = [
            cls.read_condition(cond) for cond in or_list
        ]
        if 'other_cond' in condition_dict:
            condition.other_condition = cls.read_condition(
                condition_dict['other_cond'])
            condition.other_condition.type = condition_dict['other_cond'].get(
                'type')
        return condition

    def dict(self):
        result = {'type': self.type}
        result.update(self.write_condition(self))
        return result

    @classmethod
    def write_condition(cls, condition):
        result = {}
        if condition.operator is not None:
            result['op'] = condition.operator
        if condition.tag is not None:
            result['tag'] = condition.tag
            result['index'] = condition.index
        if condition.values:
            result['values'] = condition.values
        if condition.and_conditions:
            result['and'] = []
            for and_condition in condition.and_conditions:
                result['and'].append(cls.write_condition(and_condition))
        if condition.or_conditions:
            result['or'] = []
            for or_condition in condition.or_conditions:
                result['or'].append(cls.write_condition(or_condition))
        if condition.other_condition:
            result['other_cond'] = condition.other_condition.dict()
        return result

    def to_string(self, dict_info):
        """Return a condition readable as part of a sentence."""
        result = ''
        if self.and_conditions:
            return ' and '.join(
                cond.to_string(dict_info) for cond in self.and_conditions)
        if self.or_conditions:
            return ' or '.join(
                cond.to_string(dict_info) for cond in self.or_conditions)
        if self.tag is not None:
            if dict_info and self.tag in dict_info:
                result = dict_info[self.tag]['name']
            else:
                result = self.tag
            if self.index:
                result += '[{}]'.format(self.index)
        if self.operator == '+':
            result += ' exists'
        elif self.operator == '++':
            result += ' exists and has a value'
        elif self.operator == '=>':
            tag_value = int(self.values[0])
            result += ' points to ' + tag_name_from_id(tag_value, dict_info)
        elif self.operator == '-':
            result += ' is not present'
        elif self.operator == '=':
            result += ' is equal to '
            values = ['"' + value + '"' for value in self.values]
            if len(values) > 1:
                result += ', '.join(values[:-1]) + ' or '
            result += values[-1]
        elif self.operator == '!=':
            result += ' is not equal to '
            values = ['"' + value + '"' for value in self.values]
            if len(values) > 1:
                result += ', '.join(values[:-1]) + ' and '
            result += values[-1]
        elif self.operator == '<':
            result += ' is less than ' + self.values[0]
        elif self.operator == '>':
            result += ' is greater than ' + self.values[0]
        return result



"""
Chapter6Reader collects DICOM Data Element information.
The information is taken from DICOM dictionary (PS3.6) in docbook format
as provided by ACR NEMA.
"""
from dicom_validator.spec_reader.spec_reader import (
    SpecReader, SpecReaderParseError
)


class Part6Reader(SpecReader):
    """Reads information from PS3.4 in docbook format."""

    def __init__(self, spec_dir):
        super(Part6Reader, self).__init__(spec_dir)
        self.part_nr = 6
        self._uids = None
        self._data_elements = None

    def data_elements(self):
        """Return the information about registered DICOM data elements.

        The return value is a dict with the the tag ID (group/element tuple)
        as key.
        See data_element() for the contained value.
        """
        if self._data_elements is None:
            self._read_element_table()
        return self._data_elements

    def data_element(self, tag_id):
        """Return the information about the specified tag.

        Arguments:
            tag_id: The tag ID as string in format (####,####)
        The return value is a dict with the the tag ID (group/element tuple)
        as key.
        The values of the retruned dict are dicts with the following entries:
            'name': The human readable tag name
            'vr': The tag value representation (e.g. 'PN')
            'vm': The tag multiplicity (e.g. '1-N')
            'prop': Additional properties, like 'RET' for retired
        """
        return self.data_elements().get(tag_id)

    def _read_element_table(self):
        self._data_elements = {}
        table = self._find(self._get_doc_root(),
                           ['chapter[@label="6"]', 'table', 'tbody'])
        if table is None:
            raise SpecReaderParseError(
                'Registry of DICOM Data Elements not found in PS3.6')
        row_nodes = self._findall(table, ['tr'])
        attrib_indexes = [1, 3, 4, 5]
        for row_node in row_nodes:
            column_nodes = self._findall(row_node, ['td'])
            if len(column_nodes) == 6:
                tag_id = self._find_text(column_nodes[0])
                if tag_id:
                    tag_attributes = [self._find_text(column_nodes[i])
                                      for i in attrib_indexes]
                    if tag_attributes is not None:
                        self._data_elements[tag_id] = {
                            'name': tag_attributes[0],
                            'vr': tag_attributes[1],
                            'vm': tag_attributes[2],
                            'prop': tag_attributes[3]
                        }

    def uids(self, uid_type):
        """Return a dict of UID values (keys) and names for the given UID type.
        """
        return self._get_uids().get(uid_type, {})

    def all_uids(self):
        """Return a dict of UID types with UID value/name dicts for the
        given UID type as value.
        """
        return self._get_uids()

    def sop_class_uids(self):
        """Return a dict of SOP Class UID values (keys) and names."""
        return self.uids('SOP Class')

    def sop_class_name(self, uid):
        """Return the name of SOP Class corresponding to the given UID."""
        return self.uids('SOP Class').get(uid)

    def sop_class_uid(self, sop_class_name):
        """Return the name of SOP Class corresponding to the given UID."""
        for uid, name in self.sop_class_uids().items():
            if name == sop_class_name:
                return uid

    def _get_uids(self):
        if self._uids is None:
            self._uids = {}
            table = self._find(self._get_doc_root(),
                               ['chapter[@label="A"]', 'table', 'tbody'])
            if table is None:
                raise SpecReaderParseError(
                    'Registry of DICOM Unique Identifiers not found in PS3.6')

            row_nodes = self._findall(table, ['tr'])
            for row_node in row_nodes:
                column_nodes = self._findall(row_node, ['td'])
                nr_columns = len(column_nodes)
                if nr_columns in (4, 5):
                    # columns are UID Value, UID Name, UID Keyword (only
                    # since 2020d), UID Type and Part
                    uid_attributes = [self._find_text(column_nodes[i])
                                      for i in range(nr_columns - 1)]
                    if uid_attributes is not None:
                        uid_type = uid_attributes[nr_columns - 2]
                        # in PS3.6 xml there are multiple zero width (U+200B)
                        # spaces inside the UIDs
                        # we remove them hoping this is the only such problem
                        uid_value = self.cleaned_value(uid_attributes[0])
                        self._uids.setdefault(uid_type, {})[
                            uid_value] = self.cleaned_value(uid_attributes[1])
        return self._uids



import os
import unittest

import pyfakefs.fake_filesystem_unittest

from dicom_validator.spec_reader.spec_reader import (
    SpecReader, SpecReaderFileError
)


class ReadSpecTest(pyfakefs.fake_filesystem_unittest.TestCase):
    def setUp(self):
        super(ReadSpecTest, self).setUp()
        self.setUpPyfakefs()

    def test_missing_path(self):
        spec_path = '/var/dicom/specs'
        self.assertRaises(OSError, SpecReader, spec_path)

    def test_missing_doc_files(self):
        spec_path = '/var/dicom/specs'
        os.makedirs(spec_path)
        self.fs.create_file(os.path.join('notadoc.xml'))
        self.assertRaises(SpecReaderFileError, SpecReader, spec_path)

    def test_existing_doc_files(self):
        spec_path = '/var/dicom/specs'
        os.makedirs(spec_path)
        self.fs.create_file(os.path.join(spec_path, 'part03.xml'))
        self.assertTrue(SpecReader(spec_path))

    def test_cleaned_uid(self):
        orig_value = '1.2.840.10008.5.' \
                     '\u200b1.\u200b4.\u200b1.\u200b1.\u200b88.\u200b72'
        self.assertEqual('1.2.840.10008.5.1.4.1.1.88.72',
                         SpecReader.cleaned_value(orig_value))

        if __name__ == '__main__':
            unittest.main()



import logging
import os
import sys

from pydicom import dcmread
from pydicom.errors import InvalidDicomError

from dicom_validator.validator.iod_validator import IODValidator


class DicomFileValidator(object):
    def __init__(self, iod_info, module_info, dict_info=None,
                 log_level=logging.INFO):
        self._module_info = module_info
        self._iod_info = iod_info
        self._dict_info = dict_info
        self.logger = logging.getLogger()
        self.logger.level = log_level
        if not self.logger.handlers:
            self.logger.addHandler(logging.StreamHandler(sys.stdout))

    def validate(self, path):
        errors = {}
        if not os.path.exists(path):
            errors.update({path: {'fatal': 'File missing'}})
            self.logger.warning('\n"%s" does not exist - skipping', path)
        else:
            if os.path.isdir(path):
                errors.update(self.validate_dir(path))
            else:
                errors.update(self.validate_file(path))
        return errors

    def validate_dir(self, dir_path):
        errors = {}
        for root, _, names in os.walk(dir_path):
            for name in names:
                errors.update(self.validate(os.path.join(root, name)))
        return errors

    def validate_file(self, file_path):
        self.logger.info('\nProcessing DICOM file "%s"', file_path)
        try:
            data_set = dcmread(file_path, defer_size=1024)
        except InvalidDicomError:
            return {file_path: {'fatal': 'Invalid DICOM file'}}
        return {
            file_path: IODValidator(data_set, self._iod_info,
                                    self._module_info, self._dict_info,
                                    self.logger.level).validate()
        }



import os
import unittest

import pyfakefs.fake_filesystem_unittest

from dicom_validator.spec_reader.part4_reader import Part4Reader
from dicom_validator.spec_reader.spec_reader import (
    SpecReaderLookupError, SpecReaderParseError
)
from dicom_validator.tests.test_utils import spec_fixture_path


class Part4ReaderTest(pyfakefs.fake_filesystem_unittest.TestCase):
    doc_contents = None

    @classmethod
    def setUpClass(cls):
        with open(os.path.join(spec_fixture_path(),
                               'part04.xml'), 'rb') as spec_file:
            cls.doc_contents = spec_file.read()

    def setUp(self):
        super(Part4ReaderTest, self).setUp()
        self.setUpPyfakefs()
        spec_path = os.path.join('dicom', 'specs')
        part4_path = os.path.join(spec_path, 'part04.xml')
        self.fs.create_file(part4_path, contents=self.doc_contents)
        self.reader = Part4Reader(spec_path)

    def test_read_incomplete_doc_file(self):
        spec_path = '/var/dicom/specs'
        os.makedirs(spec_path)
        self.fs.create_file(
            os.path.join(spec_path, 'part04.xml'),
            contents='<book xmlns="http://docbook.org/ns/docbook">\n</book>')
        reader = Part4Reader(spec_path)
        self.assertRaises(SpecReaderParseError,
                          reader.iod_chapter, '1.2.840.10008.5.1.4.1.1.2')

    def test_sop_class_lookup(self):
        self.assertRaises(SpecReaderLookupError,
                          self.reader.iod_chapter, '1.1.1.1')
        iod_chapter = self.reader.iod_chapter(
            sop_class_uid='1.2.840.10008.5.1.4.1.1.2')
        self.assertEqual('A.3', iod_chapter)

    def test_uids_for_chapter(self):
        iod_chapters = self.reader.iod_chapters()
        self.assertIn('A.3', iod_chapters)
        self.assertEqual(1, len(iod_chapters['A.3']))
        self.assertIn('A.26', iod_chapters)
        self.assertIn('1.2.840.10008.5.1.4.1.1.2', iod_chapters['A.3'])
        self.assertIn('A.26', iod_chapters)
        self.assertEqual(2, len(iod_chapters['A.26']))
        self.assertIn('1.2.840.10008.5.1.4.1.1.1.1', iod_chapters['A.26'])
        self.assertIn('1.2.840.10008.5.1.4.1.1.1.1.1', iod_chapters['A.26'])

    def test_secondary_capture_fix(self):
        iod_chapter = self.reader.iod_chapter(
            sop_class_uid='1.2.840.10008.5.1.4.1.1.7')
        self.assertEqual('A.8.1', iod_chapter)


if __name__ == '__main__':
    unittest.main()



import os

CURRENT_REVISION = '2021d'


def fixture_path():
    return os.path.join(os.path.dirname(__file__), 'fixtures')


def spec_fixture_path():
    return os.path.join(fixture_path(), CURRENT_REVISION, 'docbook')


def json_fixture_path():
    return os.path.join(fixture_path(), CURRENT_REVISION, 'json')


def dicom_fixture_path():
    return os.path.join(fixture_path(), 'dicom')



"""
SpecReader reads information from DICOM standard files in docbook format as
provided by ACR NEMA.
"""
import os

import xml.etree.ElementTree as ElementTree


class SpecReaderError(Exception):
    pass


class SpecReaderFileError(SpecReaderError):
    pass


class SpecReaderParseError(SpecReaderError):
    pass


class SpecReaderLookupError(SpecReaderError):
    pass


class SpecReader(object):
    docbook_ns = '{http://docbook.org/ns/docbook}'

    def __init__(self, spec_dir):
        self.spec_dir = spec_dir
        self.part_nr = 0
        document_files = os.listdir(self.spec_dir)
        if not document_files:
            raise SpecReaderFileError(
                u'Missing docbook files in {}'.format(self.spec_dir))
        self._doc_trees = {}

    def _get_doc_tree(self):
        if self.part_nr not in self._doc_trees:
            doc_name = 'part{:02}.xml'.format(self.part_nr)
            document_files = os.listdir(self.spec_dir)
            if doc_name not in document_files:
                raise SpecReaderFileError(
                    u'Missing docbook file {} in {}'.format(
                        doc_name, self.spec_dir))
            try:
                self._doc_trees[self.part_nr] = ElementTree.parse(
                    os.path.join(self.spec_dir, doc_name))
            except ElementTree.ParseError:
                raise SpecReaderFileError(
                    u'Parse error in docbook file {} in {}'.format(
                        doc_name, self.spec_dir))
        return self._doc_trees.get(self.part_nr)

    def _get_doc_root(self):
        doc_tree = self._get_doc_tree()
        if doc_tree:
            return doc_tree.getroot()

    def _find(self, node, elements):
        search_string = '/'.join(
            [self.docbook_ns + element for element in elements])
        if node is not None:
            return node.find(search_string)

    def _findall(self, node, elements):
        search_string = '/'.join(
            [self.docbook_ns + element for element in elements])
        return node.findall(search_string)

    def _find_text(self, node):
        try:
            para_node = self._find(node, ['para'])
            text_parts = [text.strip() for text in para_node.itertext() if
                          text.strip()]
            return ' '.join(text_parts) if text_parts else ''
        except AttributeError:
            return ''

    @staticmethod
    def cleaned_value(value):
        return value.replace(u'\u200B', '')

    @staticmethod
    def _find_all_text(node):
        text_parts = [text.strip() for text in node.itertext() if text.strip()]
        return ' '.join(text_parts) if text_parts else ''



import argparse
import logging
import os

from dicom_validator.spec_reader.edition_reader import EditionReader
from dicom_validator.validator.dicom_file_validator import DicomFileValidator


def validate(args, base_path):
    json_path = os.path.join(base_path, 'json')
    dict_info = EditionReader.load_dict_info(json_path)
    iod_info = EditionReader.load_iod_info(json_path)
    module_info = EditionReader.load_module_info(json_path)
    log_level = logging.DEBUG if args.verbose else logging.INFO
    validator = DicomFileValidator(iod_info, module_info, dict_info, log_level)
    error_nr = 0
    for dicom_path in args.dicomfiles:
        error_nr += sum(len(error) for error in
                        list(validator.validate(dicom_path).values()))
    return error_nr


def main(args=None):
    parser = argparse.ArgumentParser(
        description='Validates DICOM file IODs')
    parser.add_argument('dicomfiles',
                        help='Path(s) of DICOM files or directories '
                             'to validate',
                        nargs='+')
    parser.add_argument('--standard-path', '-src',
                        help='Base path with the DICOM specs in docbook '
                             'and json format',
                        default=os.path.join(os.path.expanduser("~"),
                                             'dicom-validator'))
    parser.add_argument('--revision', '-r',
                        help='Standard revision (e.g. "2014c"), year of '
                             'revision, "current" or "local" (latest '
                             'locally installed)',
                        default='current')
    parser.add_argument('--verbose', '-v', action='store_true',
                        help='Outputs diagnostic information')
    args = parser.parse_args(args)

    edition_reader = EditionReader(args.standard_path)
    destination = edition_reader.get_revision(args.revision)
    if destination is None:
        print(
            'Failed to get DICOM edition {} - aborting'.format(args.revision))
        return 1

    return validate(args, destination)


if __name__ == '__main__':
    exit(main())



#!/usr/bin/env python
import os

from setuptools import setup, find_packages

from dicom_validator import __version__

EXTRA = {}

BASE_PATH = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(BASE_PATH, 'README.md')) as f:
    long_description = f.read()


setup(
    name="dicom-validator",
    packages=find_packages(),
    include_package_data=True,
    version=__version__,
    install_requires=['pydicom'],
    description="Python DICOM tools using input from DICOM specs in docbook format",
    author="mrbean-bremen",
    author_email="hansemrbean@googlemail.com",
    url="http://github.com/pydicom/dicom-validator",
    keywords="dicom python",
    entry_points={
        'console_scripts': [
            'validate_iods=dicom_validator.validate_iods:main',
            'dump_dcm_info=dicom_validator.dump_dcm_info:main'
        ]
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Intended Audience :: Healthcare Industry",
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: POSIX :: Linux",
        'Operating System :: MacOS',
        "Operating System :: Microsoft :: Windows",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
    **EXTRA
)



import os
import unittest

import pyfakefs.fake_filesystem_unittest

from dicom_validator.spec_reader.part6_reader import Part6Reader
from dicom_validator.tests.test_utils import spec_fixture_path


class Part6ReaderTest(pyfakefs.fake_filesystem_unittest.TestCase):
    doc_contents = None

    @classmethod
    def setUpClass(cls):
        with open(os.path.join(spec_fixture_path(), 'part06.xml'),
                  'rb') as spec_file:
            cls.doc_contents = spec_file.read()

    def setUp(self):
        super(Part6ReaderTest, self).setUp()
        self.setUpPyfakefs()
        spec_path = os.path.join('dicom', 'docbook')
        part6_path = os.path.join(spec_path, 'part06.xml')
        self.fs.create_file(part6_path, contents=self.doc_contents)
        self.reader = Part6Reader(spec_path)

    def test_undefined_id(self):
        self.assertIsNone(self.reader.data_element('(0011,0011)'))

    def test_data_element(self):
        element = self.reader.data_element('(0008,0005)')
        self.assertIsNotNone(element)
        self.assertEqual('Specific Character Set', element['name'])
        self.assertEqual('CS', element['vr'])
        self.assertEqual('1-n', element['vm'])

    def test_data_elements(self):
        elements = self.reader.data_elements()
        self.assertEqual(8, len(elements))

    def test_sop_class_uids(self):
        sop_class_uids = self.reader.sop_class_uids()
        self.assertEqual(3, len(sop_class_uids))
        self.assertIn('1.2.840.10008.1.1', sop_class_uids)
        self.assertEqual('Verification SOP Class',
                         sop_class_uids['1.2.840.10008.1.1'])

    def test_uid_type(self):
        xfer_syntax_uids = self.reader.uids('Transfer Syntax')
        self.assertEqual(2, len(xfer_syntax_uids))
        self.assertIn('1.2.840.10008.1.2.4.80', xfer_syntax_uids)
        self.assertEqual('JPEG-LS Lossless Image Compression',
                         xfer_syntax_uids['1.2.840.10008.1.2.4.80'])

    def test_all_uids(self):
        uids = self.reader.all_uids()
        self.assertEqual(2, len(uids))
        self.assertIn('Transfer Syntax', uids)
        self.assertIn('SOP Class', uids)
        uid_nr = sum([len(uid_dict) for uid_dict in uids.values()])
        self.assertEqual(5, uid_nr)

    def test_sop_class_name(self):
        self.assertEqual(
            'Enhanced US Volume Storage',
            self.reader.sop_class_name('1.2.840.10008.5.1.4.1.1.6.2'))

    def test_sop_class_uid(self):
        self.assertEqual('1.2.840.10008.5.1.4.1.1.2',
                         self.reader.sop_class_uid('CT Image Storage'))


if __name__ == '__main__':
    unittest.main()



import os
import unittest

import pyfakefs.fake_filesystem_unittest

from dicom_validator.spec_reader.part3_reader import Part3Reader
from dicom_validator.spec_reader.spec_reader import (
    SpecReaderLookupError, SpecReaderParseError, SpecReaderFileError
)
from dicom_validator.tests.test_utils import spec_fixture_path


class ReadPart3Test(pyfakefs.fake_filesystem_unittest.TestCase):
    doc_contents = None

    @classmethod
    def setUpClass(cls):
        with open(os.path.join(spec_fixture_path(),
                               'part03.xml'), 'rb') as spec_file:
            cls.doc_contents = spec_file.read()

    def setUp(self):
        super(ReadPart3Test, self).setUp()
        self.setUpPyfakefs()
        spec_path = os.path.join('dicom', 'specs')
        part3_path = os.path.join(spec_path, 'part03.xml')
        self.fs.create_file(part3_path, contents=self.doc_contents)
        self.reader = Part3Reader(spec_path)

    def test_read_empty_doc_file(self):
        spec_path = '/var/dicom/specs'
        os.makedirs(spec_path)
        self.fs.create_file(os.path.join(spec_path, 'part03.xml'))
        spec_reader = Part3Reader(spec_path)
        self.assertRaises(SpecReaderFileError,
                          spec_reader.iod_description, 'A.16')

    def test_read_invalid_doc_file(self):
        spec_path = '/var/dicom/specs'
        os.makedirs(spec_path)
        self.fs.create_file(os.path.join(spec_path, 'part03.xml'),
                            contents='Not an xml')
        spec_reader = Part3Reader(spec_path)
        self.assertRaises(SpecReaderFileError,
                          spec_reader.iod_description, 'A.6')

    def test_read_incomplete_doc_file(self):
        spec_path = '/var/dicom/specs'
        os.makedirs(spec_path)
        self.fs.create_file(os.path.join(
            spec_path, 'part03.xml'),
            contents='<book xmlns="http://docbook.org/ns/docbook">\n</book>')
        reader = Part3Reader(spec_path)
        self.assertRaises(SpecReaderParseError, reader.iod_description, 'A.6')

    def test_lookup_sop_class(self):
        self.assertRaises(SpecReaderLookupError,
                          self.reader.iod_description, 'A.0')
        description = self.reader.iod_description(chapter='A.3')
        self.assertIsNotNone(description)
        self.assertTrue('title' in description)
        self.assertEqual(description['title'], 'Computed Tomography Image IOD')

    def test_get_iod_modules(self):
        description = self.reader.iod_description(chapter='A.38.1')
        self.assertIn('modules', description)
        modules = description['modules']
        self.assertEqual(27, len(modules))
        self.assertIn('General Equipment', modules)
        module = modules['General Equipment']
        self.assertEqual('C.7.5.1', module['ref'])
        self.assertEqual('M', module['use'])

    def test_optional_iod_module(self):
        description = self.reader.iod_description(chapter='A.38.1')
        self.assertIn('modules', description)
        modules = description['modules']
        self.assertIn('Clinical Trial Subject', modules)
        module = modules['Clinical Trial Subject']
        self.assertEqual('C.7.1.3', module['ref'])
        self.assertEqual('U', module['use'])

    def test_iod_descriptions(self):
        descriptions = self.reader.iod_descriptions()
        self.assertEqual(4, len(descriptions))
        self.assertIn('A.3', descriptions)
        self.assertIn('A.18', descriptions)
        self.assertIn('A.38.1', descriptions)

    def test_module_description(self):
        self.assertRaises(SpecReaderLookupError,
                          self.reader.module_description, 'C.9.9.9')
        description = self.reader.module_description('C.7.1.3')
        self.assertEqual(9, len(description))
        self.assertIn('(0012,0031)', description)
        self.assertEqual('Clinical Trial Site Name',
                         description['(0012,0031)']['name'])
        self.assertEqual('2', description['(0012,0031)']['type'])

    def test_sequence_inside_module_description(self):
        description = self.reader.module_description('C.7.2.3')
        self.assertEqual(3, len(description))
        self.assertIn('(0012,0083)', description)
        self.assertIn('items', description['(0012,0083)'])
        sequence_description = description['(0012,0083)']['items']
        self.assertEqual(3, len(sequence_description))
        self.assertIn('(0012,0020)', sequence_description)
        self.assertEqual('Clinical Trial Protocol ID',
                         sequence_description['(0012,0020)']['name'])
        self.assertEqual('1C', sequence_description['(0012,0020)']['type'])

    def test_referenced_macro(self):
        # module has 2 directly included attributes
        # and 21 attribute in referenced table
        description = self.reader.module_description('C.7.6.3')
        self.assertEqual(3, len(description))
        self.assertIn('(0028,7FE0)', description)
        self.assertIn('include', description)
        self.assertIn('C.7-11b', description['include'])
        description = self.reader.module_description('C.7-11b')
        self.assertEqual(21, len(description))
        self.assertIn('(7FE0,0010)', description)

    def test_module_descriptions(self):
        descriptions = self.reader.module_descriptions()
        # 42 modules from 3 classes (20/23/26) with overlapping
        # common modules + 27 referenced macros
        self.assertEqual(69, len(descriptions))


if __name__ == '__main__':
    unittest.main()



import json
import os
import unittest

from dicom_validator.spec_reader.condition_parser import ConditionParser
from dicom_validator.tests.test_utils import json_fixture_path
from dicom_validator.spec_reader.edition_reader import EditionReader


class ConditionParserTest(unittest.TestCase):
    """Tests ConditionParser by testing different kinds of condition strings
    as taken from DICOM standard.
    """
    dict_info = None

    @classmethod
    def setUpClass(cls):
        with open(os.path.join(json_fixture_path(),
                               EditionReader.dict_info_json)) as info_file:
            cls.dict_info = json.load(info_file)

    def setUp(self):
        super(ConditionParserTest, self).setUp()
        self.parser = ConditionParser(self.dict_info)


class InvalidConditionParserTest(ConditionParserTest):
    def test_ignore_invalid_condition(self):
        result = self.parser.parse('')
        self.assertIsNone(result.tag)
        self.assertEqual('U', result.type)

    def test_ignore_uncheckable_tag_condition(self):
        result = self.parser.parse(
            'Required if Numeric Value (0040,A30A) has insufficient '
            'precision to represent the value as a string.')
        self.assertEqual('U', result.type)
        self.assertIsNone(result.tag)

    def test_ignore_condition_without_tag(self):
        result = self.parser.parse(
            'Required if present and consistent '
            'in the contributing SOP Instances. ')
        self.assertEqual('U', result.type)


class SimpleConditionParserTest(ConditionParserTest):
    def test_not_present(self):
        result = self.parser.parse(
            'Required if VOI LUT Sequence (0028,3010) is not present.')
        self.assertEqual('MN', result.type)
        self.assertEqual('(0028,3010)', result.tag)
        self.assertEqual('-', result.operator)
        self.assertIsNone(result.values)

    def test_operator_in_tag(self):
        result = self.parser.parse(
            'Required if Fractional Channel Display Scale '
            '(003A,0247) is not present')
        self.assertEqual('MN', result.type)
        self.assertEqual('(003A,0247)', result.tag)
        self.assertEqual('-', result.operator)
        self.assertIsNone(result.values)

    def test_is_present(self):
        result = self.parser.parse(
            'Required if Bounding Box Top Left Hand Corner '
            '(0070,0010) is present.')
        self.assertEqual('MN', result.type)
        self.assertEqual('(0070,0010)', result.tag)
        self.assertEqual('+', result.operator)
        self.assertIsNone(result.values)

    def test_is_present_with_value(self):
        result = self.parser.parse(
            'Required if Responsible Person is present and has a value.')
        self.assertEqual('MN', result.type)
        self.assertEqual('(0010,2297)', result.tag)
        self.assertEqual('++', result.operator)
        self.assertIsNone(result.values)

    def test_is_present_tag_name_with_digit(self):
        result = self.parser.parse(
            'Required if 3D Mating Point (0068,64C0) is present.')
        self.assertEqual('MN', result.type)
        self.assertEqual('(0068,64C0)', result.tag)
        self.assertEqual('+', result.operator)
        self.assertIsNone(result.values)

    def test_not_sent(self):
        result = self.parser.parse(
            'Required if Anatomic Region Modifier Sequence '
            '(0008,2220) is not sent. ')
        self.assertEqual('MN', result.type)
        self.assertEqual('(0008,2220)', result.tag)
        self.assertEqual('-', result.operator)
        self.assertIsNone(result.values)

    def test_shall_be_condition_with_absent_tag(self):
        result = self.parser.parse(
            'Some Stuff. Shall be present if Clinical Trial Subject Reading ID'
            ' (0012,0042) is absent. May be present otherwise.')
        self.assertEqual('MU', result.type)
        self.assertEqual('(0012,0042)', result.tag)
        self.assertEqual(0, result.index)
        self.assertEqual('-', result.operator)
        self.assertIsNone(result.values)


class ValueConditionParserTest(ConditionParserTest):
    def test_equality_tag(self):
        result = self.parser.parse(
            'C - Required if Modality (0008,0060) = IVUS')
        self.assertEqual('MN', result.type)
        self.assertEqual('(0008,0060)', result.tag)
        self.assertEqual(0, result.index)
        self.assertEqual('=', result.operator)
        self.assertEqual(['IVUS'], result.values)

    def test_equality_tag_without_tag_id(self):
        result = self.parser.parse('C - Required if Modality = IVUS')
        self.assertEqual('MN', result.type)
        self.assertEqual('(0008,0060)', result.tag)
        self.assertEqual(0, result.index)
        self.assertEqual('=', result.operator)
        self.assertEqual(['IVUS'], result.values)

    def test_multiple_values_and_index(self):
        result = self.parser.parse(
            'C - Required if Image Type (0008,0008) Value 3 '
            'is GATED, GATED TOMO, or RECON GATED TOMO')
        self.assertEqual('MN', result.type)
        self.assertEqual('(0008,0008)', result.tag)
        self.assertEqual(2, result.index)
        self.assertEqual('=', result.operator)
        self.assertEqual(['GATED', 'GATED TOMO', 'RECON GATED TOMO'],
                         result.values)

    def test_comma_before_value(self):
        result = self.parser.parse(
            'Required if Series Type (0054,1000), Value 2 is REPROJECTION.')
        self.assertEqual('MN', result.type)
        self.assertEqual('(0054,1000)', result.tag)
        self.assertEqual(1, result.index)
        self.assertEqual('=', result.operator)
        self.assertEqual(['REPROJECTION'], result.values)

    def test_may_be_present_otherwise(self):
        result = self.parser.parse(
            'C - Required if Image Type (0008,0008) Value 1 equals ORIGINAL.'
            ' May be present otherwise.')
        self.assertEqual('MU', result.type)
        self.assertEqual('(0008,0008)', result.tag)
        self.assertEqual(0, result.index)
        self.assertEqual('=', result.operator)
        self.assertEqual(['ORIGINAL'], result.values)

    def test_greater_operator(self):
        result = self.parser.parse(
            'C - Required if Number of Frames is greater than 1')
        self.assertEqual('MN', result.type)
        self.assertEqual('(0028,0008)', result.tag)
        self.assertEqual(0, result.index)
        self.assertEqual('>', result.operator)
        self.assertEqual(['1'], result.values)

    def test_value_greater_than_operator(self):
        result = self.parser.parse(
            'Required if Samples per Pixel '
            '(0028,0002) has a value greater than 1')
        self.assertEqual('MN', result.type)
        self.assertEqual('(0028,0002)', result.tag)
        self.assertEqual(0, result.index)
        self.assertEqual('>', result.operator)
        self.assertEqual(['1'], result.values)

    def test_tag_ids_as_values(self):
        result = self.parser.parse(
            'C - Required if Frame Increment Pointer (0028,0009) '
            'is Frame Time (0018,1063) or Frame Time Vector (0018,1065)')
        self.assertEqual('MN', result.type)
        self.assertEqual('(0028,0009)', result.tag)
        self.assertEqual(0, result.index)
        self.assertEqual('=', result.operator)
        self.assertEqual(
            ['Frame Time (0018,1063)', 'Frame Time Vector (0018,1065)'],
            result.values)

    def test_has_a_value_of(self):
        result = self.parser.parse(
            'Required if Pixel Presentation '
            '(0008,9205) has a value of TRUE_COLOR.')
        self.assertEqual('MN', result.type)
        self.assertEqual('(0008,9205)', result.tag)
        self.assertEqual('=', result.operator)
        self.assertEqual(['TRUE_COLOR'], result.values)

    def test_at_the_image_level_equals(self):
        result = self.parser.parse(
            '"Required if Pixel Presentation (0008,9205) at the image level '
            'equals COLOR or MIXED.')
        self.assertEqual('MN', result.type)
        self.assertEqual('=', result.operator)
        self.assertEqual(['COLOR', 'MIXED'], result.values)

    def test_is_with_colon(self):
        result = self.parser.parse(
            'Required if Image Type '
            '(0008,0008) Value 3 is: WHOLE BODY or STATIC.')
        self.assertEqual('MN', result.type)
        self.assertEqual('(0008,0008)', result.tag)
        self.assertEqual('=', result.operator)
        self.assertEqual(2, result.index)
        self.assertEqual(['WHOLE BODY', 'STATIC'], result.values)

    def test_remove_apostrophes(self):
        result = self.parser.parse(
            'Required if Lossy Image Compression (0028,2110) is "01".')
        self.assertEqual('=', result.operator)
        self.assertEqual(['01'], result.values)

    def test_remove_apostrophes_from_uids(self):
        result = self.parser.parse(
            'Required if SOP Class UID (0008,0016) '
            'equals "1.2.840.10008.5.1.4.1.1.12.1.1" '
            'or "1.2.840.10008.5.1.4.1.1.12.2.1". May be present otherwise.')
        self.assertEqual('=', result.operator)
        self.assertEqual(['1.2.840.10008.5.1.4.1.1.12.1.1',
                          '1.2.840.10008.5.1.4.1.1.12.2.1'], result.values)

    def test_value_of(self):
        result = self.parser.parse(
            'Required if the value of Context Group Extension Flag '
            '(0008,010B) is "Y".')
        self.assertEqual('MN', result.type)
        self.assertEqual('=', result.operator)
        self.assertEqual(['Y'], result.values)

    def test_value_more_than(self):
        result = self.parser.parse(
            'Required if Data Point Rows '
            '(0028,9001) has a value of more than 1.')
        self.assertEqual('MN', result.type)
        self.assertEqual('(0028,9001)', result.tag)
        self.assertEqual('>', result.operator)
        self.assertEqual(['1'], result.values)

    def test_is_not_with_uid(self):
        result = self.parser.parse(
            'Required if SOP Class UID is not '
            '"1.2.840.10008.5.1.4.1.1.4.4" (Legacy Converted).')
        self.assertEqual('MN', result.type)
        self.assertEqual('(0008,0016)', result.tag)
        self.assertEqual('!=', result.operator)
        self.assertEqual(['1.2.840.10008.5.1.4.1.1.4.4'], result.values)

    def test_present_with_value(self):
        result = self.parser.parse(
            'Required if Selector Attribute VR '
            '(0072,0050) is present and the value is AS.')
        self.assertEqual('MN', result.type)
        self.assertEqual('=', result.operator)
        self.assertEqual(['AS'], result.values)

    def test_other_than(self):
        result = self.parser.parse(
            'Required if Decay Correction (0054,1102) is other than NONE.')
        self.assertEqual('MN', result.type)
        self.assertEqual('!=', result.operator)
        self.assertEqual(['NONE'], result.values)

    def test_not_equal_to(self):
        result = self.parser.parse(
            'Required if Planes in Acquisition '
            '(0018,9410) is not equal to UNDEFINED.')
        self.assertEqual('MN', result.type)
        self.assertEqual('!=', result.operator)
        self.assertEqual(['UNDEFINED'], result.values)

    def test_present_with_value_of(self):
        result = self.parser.parse(
            'Required if Partial View '
            '(0028,1350) is present with a value of YES.')
        self.assertEqual('MN', result.type)
        self.assertEqual('=', result.operator)
        self.assertEqual(['YES'], result.values)

    def test_points_to_tag(self):
        result = self.parser.parse(
            'Required if Frame Increment Pointer (0028,0009) points to '
            'Frame Label Vector (0018,2002).')
        self.assertEqual('MN', result.type)
        self.assertEqual('=>', result.operator)
        self.assertEqual(['1581058'], result.values)


class NotMandatoryConditionParserTest(ConditionParserTest):
    def test_default(self):
        result = self.parser.parse(
            'Required if Image Type '
            '(0008,0008) Value 1 is ORIGINAL. May be present otherwise.')
        self.assertEqual('MU', result.type)
        self.assertEqual('=', result.operator)
        self.assertEqual(['ORIGINAL'], result.values)

    def test_comma_instead_of_dot(self):
        result = self.parser.parse(
            'Required if Absolute Channel Display Scale '
            '(003A,0248) is not present, '
            'may be present otherwise.')
        self.assertEqual('MU', result.type)
        self.assertEqual('-', result.operator)

    def test_missing_dot(self):
        result = self.parser.parse(
            'Required if Image Type '
            '(0008,0008) Value 1 is ORIGINAL May be present otherwise.')
        self.assertEqual('MU', result.type)
        self.assertEqual('=', result.operator)
        self.assertEqual(['ORIGINAL'], result.values)


class CompositeConditionParserTest(ConditionParserTest):
    def test_and_condition(self):
        result = self.parser.parse(
            'Required if Series Type (0054,1000), Value 1 is GATED and '
            'Beat Rejection Flag (0018,1080) is Y.')
        self.assertEqual('MN', result.type)
        self.assertEqual(2, len(result.and_conditions))
        result1 = result.and_conditions[0]
        self.assertEqual('(0054,1000)', result1.tag)
        self.assertEqual('=', result1.operator)
        self.assertEqual(['GATED'], result1.values)
        result2 = result.and_conditions[1]
        self.assertEqual('(0018,1080)', result2.tag)
        self.assertEqual('=', result2.operator)
        self.assertEqual(['Y'], result2.values)

    def test_ignore_unverifyable_and_condition(self):
        result = self.parser.parse(
            'Required if Delivery Type (300A,00CE) is CONTINUATION and '
            'one or more channels of any Application Setup are omitted.')
        self.assertEqual('MN', result.type)
        self.assertEqual(0, len(result.and_conditions))
        self.assertEqual('=', result.operator)
        self.assertEqual(['CONTINUATION'], result.values)

    def test_and_without_value(self):
        result = self.parser.parse(
            'Required if Recorded Channel Sequence (3008,0130) is sent and '
            'Brachy Treatment Type (300A,0202) is not MANUAL or PDR.')
        self.assertEqual('MN', result.type)
        self.assertEqual(2, len(result.and_conditions))
        result1 = result.and_conditions[0]
        self.assertEqual('(3008,0130)', result1.tag)
        self.assertEqual('+', result1.operator)
        result2 = result.and_conditions[1]
        self.assertEqual('(300A,0202)', result2.tag)
        self.assertEqual('!=', result2.operator)
        self.assertEqual(['MANUAL', 'PDR'], result2.values)

    def test_and_with_multiple_values(self):
        result = self.parser.parse(
            'Required if Image Type (0008,0008) Value 1 is ORIGINAL or MIXED '
            'and Respiratory Motion Compensation Technique '
            '(0018,9170) equals other than NONE.')
        self.assertEqual('MN', result.type)
        self.assertEqual(2, len(result.and_conditions))
        self.assertEqual('=', result.and_conditions[0].operator)
        self.assertEqual(['ORIGINAL', 'MIXED'],
                         result.and_conditions[0].values)
        self.assertEqual('!=', result.and_conditions[1].operator)
        self.assertEqual(['NONE'], result.and_conditions[1].values)

    def test_either_or_tag_presence(self):
        result = self.parser.parse(
            "Required if either Patient's Birth Date in Alternative Calendar "
            "(0010,0033) or Patient's Alternative Death Date in Calendar "
            "(0010,0034) is present.")
        self.assertEqual('MN', result.type)
        result1 = result.or_conditions[0]
        self.assertEqual('(0010,0033)', result1.tag)
        self.assertEqual('+', result1.operator)
        result2 = result.or_conditions[1]
        self.assertEqual('(0010,0034)', result2.tag)
        self.assertEqual('+', result2.operator)

    def test_multiple_tag_absence(self):
        result = self.parser.parse(
            'Required if DICOM Media Retrieval Sequence (0040,E022), '
            'WADO Retrieval Sequence (0040,E023), WADO-RS Retrieval Sequence '
            '(0040,E025) and XDS Retrieval Sequence '
            '(0040,E024) are not present. May be present otherwise.')
        self.assertEqual('MU', result.type)
        self.assertEqual(4, len(result.and_conditions))
        for result_part in result.and_conditions:
            self.assertEqual('-', result_part.operator)

    def test_multiple_tag_absence_with_comma(self):
        result = self.parser.parse(
            'Required if DICOM Retrieval Sequence (0040,E021), '
            'WADO Retrieval Sequence (0040,E023), '
            'and WADO-RS Retrieval Sequence (0040,E025) '
            'and XDS Retrieval Sequence (0040,E024) are not present.')
        self.assertEqual('MN', result.type)
        self.assertEqual(4, len(result.and_conditions))
        for result_part in result.and_conditions:
            self.assertEqual('-', result_part.operator)

    def test_multiple_tag_presence(self):
        result = self.parser.parse(
            'Required if Selector Attribute (0072,0026) and '
            'Filter-by Operator (0072,0406) are present.')
        self.assertEqual('MN', result.type)
        self.assertEqual(2, len(result.and_conditions))
        for result_part in result.and_conditions:
            self.assertEqual('+', result_part.operator)

    def test_mixed_and_or_tag_presence(self):
        result = self.parser.parse(
            'Required if Selector Attribute (0072,0026) or Filter-by Category '
            '(0072,0402), and Filter-by Operator (0072,0406) are present.')
        self.assertEqual('MN', result.type)
        self.assertEqual(2, len(result.and_conditions))
        self.assertEqual(2, len(result.and_conditions[0].or_conditions))
        for result_part in result.and_conditions[0].or_conditions:
            self.assertEqual('+', result_part.operator)
        self.assertEqual('+', result.and_conditions[1].operator)

    def test_multi_tag_in_second_condition(self):
        result = self.parser.parse(
            'Required if Temporal Range Type (0040,A130) is present, '
            'and if Referenced Time Offsets (0040,A138) and '
            'Referenced DateTime (0040,A13A) are not present.')
        self.assertEqual('MN', result.type)
        self.assertEqual(2, len(result.and_conditions))
        self.assertEqual(2, len(result.and_conditions[1].and_conditions))
        for result_part in result.and_conditions[1].and_conditions:
            self.assertEqual('-', result_part.operator)
        self.assertEqual('+', result.and_conditions[0].operator)

    def test_is_present_with_multiple_tags(self):
        result = self.parser.parse(
            'Required if Bounding Box Top Left Hand Corner (0070,0010) '
            'or Bounding Box Bottom Right Hand Corner (0070,0011) is present.')
        self.assertEqual('MN', result.type)
        self.assertEqual(2, len(result.or_conditions))
        for result_part in result.or_conditions:
            self.assertEqual('+', result_part.operator)

    def test_multiple_tags_with_value(self):
        result = self.parser.parse(
            'Required if the value of Image Box Layout Type '
            '(0072,0304) is TILED, and the value of '
            'Image Box Tile Horizontal Dimension (0072,0306) or '
            'Image Box Tile Vertical Dimension (0072,0308) is greater than 1.')
        self.assertEqual('MN', result.type)
        self.assertEqual(2, len(result.and_conditions))
        self.assertEqual(2, len(result.and_conditions[1].or_conditions))
        for result_part in result.and_conditions[1].or_conditions:
            self.assertEqual('>', result_part.operator)
            self.assertTrue(result_part.values)
            self.assertEqual('1', result_part.values[0])

    def test_ispresent_with_value(self):
        result = self.parser.parse(
            'Required if Patient Identity Removed (0012,0062) is present and '
            'has a value of YES and De-identification Method Code Sequence '
            '(0012,0064) is not present.')
        self.assertEqual('MN', result.type)
        self.assertEqual(2, len(result.and_conditions))
        self.assertEqual('=', result.and_conditions[0].operator)
        self.assertEqual('YES', result.and_conditions[0].values[0])
        self.assertEqual('-', result.and_conditions[1].operator)

    def check_or_condition(self, result):
        self.assertEqual('MN', result.type)
        self.assertEqual(2, len(result.or_conditions))
        self.assertEqual('=', result.or_conditions[0].operator)
        self.assertEqual('PALETTE COLOR', result.or_conditions[0].values[0])
        self.assertEqual('=', result.or_conditions[1].operator)
        self.assertEqual(['COLOR', 'MIXED'], result.or_conditions[1].values)

    def test_or_condition_with_space(self):
        result = self.parser.parse(
            '"Required if Photometric Interpretation '
            '(0028,0004) has a value of PALETTE COLOR '
            'or Pixel Presentation (0008,9205) equals COLOR or MIXED.')
        self.check_or_condition(result)

    def test_or_condition_with_comma(self):
        result = self.parser.parse(
            '"Required if Photometric Interpretation '
            '(0028,0004) has a value of PALETTE COLOR, '
            'or Pixel Presentation (0008,9205) equals COLOR or MIXED.')
        self.check_or_condition(result)


class ComplicatedConditionParserTest(ConditionParserTest):
    def disabled_test_ispresent_with_value(self):
        result = self.parser.parse(
            'Required if Graphic Data (0070,0022) is "closed", '
            'that is Graphic Type (0070,0023) is CIRCLE or ELLIPSE, '
            'or Graphic Type (0070,0023) is POLYLINE or INTERPOLATED '
            'and the first data point is the same as the last data point.')
        self.assertEqual('MN', result.type)
        self.assertEqual(3, len(result.or_conditions))
        self.assertEqual('=', result.or_conditions[0].operator)
        self.assertEqual(['closed'], result.or_conditions[0].values)
        self.assertEqual('=', result.or_conditions[1].operator)
        self.assertEqual(['CIRCLE', 'ELLIPSE'], result.or_conditions[1].values)
        self.assertEqual('=', result.or_conditions[2].operator)
        self.assertEqual(['POLYLINE', 'INTERPOLATED'],
                         result.or_conditions[2].values)

    def test_other_condition(self):
        result = self.parser.parse(
            'Required if 3D Point Coordinates (0068,6590) is not present and '
            'HPGL Document Sequence (0068,62C0) is present. '
            'May be present if 3D Point Coordinates '
            '(0068,6590) is present and '
            'HPGL Document Sequence (0068,62C0) is present.')
        self.assertEqual('MC', result.type)
        self.assertEqual(2, len(result.and_conditions))
        self.assertEqual('-', result.and_conditions[0].operator)
        self.assertEqual('+', result.and_conditions[1].operator)
        other_cond = result.other_condition
        self.assertIsNotNone(other_cond)
        self.assertEqual(2, len(other_cond.and_conditions))
        self.assertEqual('+', other_cond.and_conditions[0].operator)
        self.assertEqual('(0068,6590)', other_cond.and_conditions[0].tag)
        self.assertEqual('+', other_cond.and_conditions[1].operator)
        self.assertEqual('(0068,62C0)', other_cond.and_conditions[1].tag)
